/***************************************************************************
 *   Copyright (C) 2006 by André Lindhjem <belgarat@sdf.lonestar.org>,     *
 *                         Kjetil Holien <kjetil.holien@gmail.com>,        *
 *                         Terje Risa <terje.risa@gmail.com> &             *
 *                         Øyvind Nerbråten <oyvind@nerbraten.com>         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************
 *
 * \file audio.c
 * \author André Lindhjem, Kjetil Holien, Terje Risa & Øyvind Nerbråten
 * \date 25.02.2006
 */


/**
 * Includes
 */
#include <unistd.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <fcntl.h>
#include <assert.h>
#include <stdbool.h>
#include <sys/select.h>	/* for select () */
#include <sys/time.h>	/* for timeval */

#include "common.h"		/* common project settings */
#include "audio.h"		/* common for all audio */
#include "mp3.h"		/* mp3 decoder */
#include "parsetime.h"	/* converts time-formats */
#include "report.h"		/* reporter class */
#include "snprintf/snprintf.h"

#ifdef WITH_AAC
#include "aac.h"		/* aac decoder */
#endif	/* WITH_AAC */


/**
 * Global vars
 */
static
audio_data_t *data = NULL;	/* the main data struct. this contains all information relating to the audio file and playback */

#define AUDIO_MAGIC_NUMBER 0x6B2F6EEF


/**
 * Function declarations
 */
static
bool open_file (const char *fname, audio_data_t *data);

static
audio_file_t get_file_type (const char *file);

static
bool is_new_file (audio_data_t *data);

static
void *audio_decoder (void*);

/**
 * Function definitions
 */
/**
 * Finds out the type of file and starts the aproporiate decoder
 * 
 * data must be mutex locked when this function is called.
 */
static void *audio_decoder (void* unused)
{
	audio_state_t playing;
	
	unused = NULL;

/*	while (pthread_mutex_trylock (&data->mutex) == EBUSY) { printf ("audio_decoder: retrying mutex\n"); }*/	/* get mutex - we work on "data" */
	pthread_mutex_lock (&data->mutex);	/* get mutex - we work on "data" */
	if (DEBUG_THREADS) printf ("audio_decoder has mutex\n");
	
	assert (data->magic == AUDIO_MAGIC_NUMBER);	/* check magic number */

	if (DEBUG_AUDIO) report ("playing file. . ", REP_DEBUG);

	switch (get_file_type(data->file_name))	/* finds out the filetype */
	{
#ifdef WITH_AAC	/* do we build with or without AAC support? */
		case (AUDIO_FILE_AAC):
			if (DEBUG_AUDIO) report ("audio.c: audio-file is aac", REP_DEBUG);

			if ((is_new_file (data)) &&
				(data->prev_file_type != AUDIO_FILE_AAC))	/* last file_type was NOT aac */
			{
				if (DEBUG_AUDIO) report ("the last filetype was NOT aac", REP_DEBUG);
				mp3_close (data);	/* close decoder and start a new one */
				aac_init (data);
			}
			(void *) aac_play (data);
			break;
#endif	/* WITH_AAC */
		case (AUDIO_FILE_MP3):
			if (DEBUG_AUDIO) report ("audio.c: audio-file is mp3", REP_DEBUG);

			if ((is_new_file (data)) &&
				(data->prev_file_type != AUDIO_FILE_MP3)) /* last file_type was NOT mp3 */
			{
				if (DEBUG_AUDIO) report ("the last filetype was NOT mp3", REP_DEBUG);
				pthread_mutex_unlock (&data->mutex);	/** release mutex - we're done with "data" */
				if (DEBUG_THREADS) printf ("audio_decoder released mutex. calling mp3_init\n");
#ifdef WITH_AAC
				aac_close (data);	/* close decoder and start a new one */
#endif	/* WITH_AAC */
				mp3_init (data);
/*				while (pthread_mutex_trylock (&data->mutex) == EBUSY) { printf ("audio_decoder: before mp3_play\n"); }*/	/* get mutex - we work on "data" */
				pthread_mutex_lock (&data->mutex);
				if (DEBUG_THREADS) printf ("audio_decoder after mp3_init has mutex\n");
			}
			pthread_mutex_unlock (&data->mutex);	/* release mutex - we're done with "data" */
			if (DEBUG_THREADS) printf ("audio_decoder released mutex before mp3_play\n");
			mp3_play (data);
/*			while (pthread_mutex_trylock (&data->mutex) == EBUSY) { printf ("audio_decoder: after mp3_play\n"); }*/	/* get mutex - we work on "data" */
			pthread_mutex_lock (&data->mutex);
			if (DEBUG_THREADS) printf ("audio_decoder has mutex after mp3_play\n");
			break;
		case (AUDIO_FILE_WAV):
			report ("wav parsing isn't implemented yet", REP_MESSAGE);
			break;
		case (AUDIO_FILE_UNKNOWN):
		default:		/* This should never happen! */
			assert (0);	/* Report where we are and exit the program */
			break;
	}

	if (DEBUG_AUDIO) report ("start_decoder thread exiting", REP_DEBUG);

	playing = data->state;

	data->is_playing = false;				/* we're not playing anymore */
	data->state = AUDIO_STATE_STOP;			/* settings state to stopped */

	data->do_callback_done = true; 

	pthread_cond_signal (&data->finished_cond);	/* signal to audio_block () that we are finished playing a segment */



	pthread_mutex_unlock (&data->mutex);	/* release mutex - we're done with "data" */
	if (DEBUG_THREADS) printf ("audio_decoder released mutex\n");

	if (DEBUG_THREADS) printf ("exiting decoder thread\n");
	pthread_exit (0);
}


/**
 * Plays a file from start to stop.
 * returns the duration in milliseconds or -1 if fail.
 */
unsigned
long
int audio_play (const char *fname, const char *start, const char *stop)
{
	unsigned long int ret;
	
/*	while (pthread_mutex_trylock (&data->mutex) == EBUSY) { printf ("audio_play: \n"); sleep(1); }*/	/** get mutex - we work on "data" */
	pthread_mutex_lock (&data->mutex);
	if (DEBUG_THREADS) printf ("audio_play has mutex\n");
	assert (data->magic == AUDIO_MAGIC_NUMBER);	/* check magic number */

	if (data->state != AUDIO_STATE_STOP) {	/* make sure we are stopped */
		data->cb_error (DAISY_ERROR_AUDIO_NOT_STOPPED, "audio_play called when state is NOT stopped");
		if (DEBUG_AUDIO) report ("audio_play called when state is NOT stopped", REP_ERROR);
		ret = -1;
	}
	else {
		data->progress.seconds	= 0;	/* reset the progress */
		data->progress.fraction	= 0;

		ret	=	parse_time_get_dur(start, stop);	/* get duration in ms */
		data->start	=	parse_time_mad_time_t (start);	/* and set start- and stop-times. */
		data->stop	=	parse_time_mad_time_t (stop);
		
		if (open_file (fname, data) == false) {	/* constructing and reporting an error-message */
			char message[STRLEN];
			snprintf (message, sizeof (message), "Could not open file. (%s)", fname);
			report (message, REP_WARNING);
			return false;
		}

		data->state = AUDIO_STATE_PLAY;		/* set state to playing */
	}	
	pthread_mutex_unlock (&data->mutex);	/* release mutex - we're done with "data" */
	if (DEBUG_THREADS) printf ("audio_play unlocked mutex\n");
	return ret;	/* we return the duration of this clip */
}


/**
 * A thread function. This is run as a thread from daisyplayer.c
 * and keeps the audio-engine alive.
 */
void* audio_thread (void* unused)
{
	pthread_t audio_decoder_thread;

	audio_state_t state;
	bool playing;
	
	unused = NULL;
	
/*	while (pthread_mutex_trylock (&data->mutex) == EBUSY) { printf ("audio_thread: retrying mutex\n"); sleep (1); }*/	/* get mutex - we work on "data" */
	pthread_mutex_lock (&data->mutex);
	if (DEBUG_THREADS) printf ("audio_thread has mutex\n");
	assert (data->magic == AUDIO_MAGIC_NUMBER);	/* check magic number */

	data->is_playing = false;
	data->state = AUDIO_STATE_STOP;

	pthread_mutex_unlock (&data->mutex);	/* release mutex - we're done with "data" */
	if (DEBUG_THREADS) printf ("audio_thread released mutex\n");
	while (1)
	{
/*		while (pthread_mutex_trylock (&data->mutex) == EBUSY) { printf ("audio_thread: loop. retrying mutex\n"); sleep (1); }*/	/* get mutex - we work on "data" */
		pthread_mutex_lock (&data->mutex);
		if (DEBUG_THREADS) printf ("audio_thread loop has mutex\n");
		assert (data->magic == AUDIO_MAGIC_NUMBER);
		state	= data->state;
		playing	= data->is_playing;

		if (data->do_callback_done) {
			pthread_mutex_unlock (&data->mutex);	/* release mutex - we're done with "data" */
			if (DEBUG_THREADS) printf ("audio_thread loop released mutex (running callback)\n");
			data->cb_playing_done (data->daisy);	/* calling callback function. this will let the
													   frontend know that we are done playing */
			data->do_callback_done = false;
		} else {
			pthread_mutex_unlock (&data->mutex);	/* release mutex - we're done with "data" */
			if (DEBUG_THREADS) printf ("audio_thread loop released mutex (didn't run callback)\n");
		}

		if (state == AUDIO_STATE_PLAY)
		{
			if (!playing)
			{
				if (DEBUG_AUDIO) report ("audio_thread: starting start_decoder thread", REP_DEBUG);
/*				while (pthread_mutex_trylock (&data->mutex) == EBUSY) { printf ("audio_thread: playing\n"); }*/	/* get mutex - we work on "data" */
				pthread_mutex_lock (&data->mutex);
				if (DEBUG_THREADS) printf ("audio_thread playing has mutex\n");
				pthread_create(&audio_decoder_thread, NULL, &audio_decoder, NULL);
				data->is_playing = true;
				pthread_mutex_unlock (&data->mutex);	/* release mutex - we're done with "data" */
				if (DEBUG_THREADS) printf ("audio_thread playing released mutex\n");
			}
		}
		if (state == AUDIO_STATE_STOP)
		{
			if (playing)
			{
				if (DEBUG_AUDIO) report ("audio_thread: stopping start_decoder thread", REP_DEBUG);
/*				while (pthread_mutex_trylock (&data->mutex) == EBUSY) { printf ("audio_thread: stop\n"); }*/	/* get mutex - we work on "data" */
				pthread_mutex_lock (&data->mutex);
				if (DEBUG_THREADS) printf ("audio_thread stop has mutex\n");
				pthread_cancel(audio_decoder_thread);	/* stopping thread */
				data->is_playing = false;				/* we're not playing anymore */
				pthread_cond_signal(&data->stopped_cond);
				pthread_mutex_unlock (&data->mutex);	/* release mutex - we're done with "data" */
				if (DEBUG_THREADS) printf ("audio_thread stop released mutex\n");
			}
		}

		usleep (500);	/* can't get select () to work. usleep seems to do the trick */

		sched_yield();	/* give up the cpu */
	}
}


/**
 * Sets the state to stop
 */
int audio_stop (void)
{
	int ret = -1;	/* return value */
	
	if (DEBUG_AUDIO) report ("stopping", REP_DEBUG);
	/*	while (pthread_mutex_trylock (&data->mutex) == EBUSY) { printf ("audio_stop: retrying mutex\n"); sleep(1); }*/	/* get mutex - we work on "data" */
	pthread_mutex_lock (&data->mutex);
	if (DEBUG_THREADS) printf ("audio_stop has mutex\n");	
	assert (data->magic == AUDIO_MAGIC_NUMBER);
	
	if (data == NULL) {	/* stop are called before we are initialized! (data doesn't exist yet!) */
		data->cb_error (DAISY_ERROR_AUDIO_DATA_IS_NULL, "audio data not initialized! (audio.c:audio_stop)");
		ret = -1;
	}
	else if (data->state == AUDIO_STATE_PLAY)
	{
		assert (data->magic == AUDIO_MAGIC_NUMBER);	/* check magic number */
		data->state = AUDIO_STATE_STOP;
		pthread_cond_wait(&data->stopped_cond, &data->mutex);
		pthread_cond_signal(&data->finished_cond);	/* signal to audio_block that we are finished */

		ret = 1;	/* stop succeeds */
	}
	else {	/* we can't stop if we aren't currently playing! */
		data->cb_error (DAISY_ERROR_AUDIO_STOPPED_WHILE_NOT_PLAYING, "stop called while we're not playing (audio.c:audio_stop)");
		ret = -1;	/* error stopping */
	}

	pthread_mutex_unlock (&data->mutex);	/* release mutex - we're done with "data" */
	return ret;
}


/**
 * Sets the state to pause
 */
int audio_pause (void)
{
	bool playing = NULL;
	audio_state_t state = AUDIO_STATE_NOOP;
	
/*	while (pthread_mutex_trylock (&data->mutex) == EBUSY) { printf ("audio_pause:\n"); }*/	/** get mutex - we work on "data" */
	pthread_mutex_lock (&data->mutex);
	if (DEBUG_THREADS) printf ("audio_stop has mutex\n");
	assert (data->magic == AUDIO_MAGIC_NUMBER);	/* check magic number */

	playing = data->is_playing;
	state = data->state;	/* copy data->state to a local variable */
	pthread_mutex_unlock (&data->mutex);	/* release mutex - we're done with "data" */

	if (playing)
	{
		if (state == AUDIO_STATE_PAUSE)
		{
			if (DEBUG_AUDIO) report ("unpausing", REP_DEBUG);
/*			while (pthread_mutex_trylock (&data->mutex) == EBUSY) { printf ("audio_pause: \n"); }*/	/** get mutex - we work on "data" */
			pthread_mutex_lock (&data->mutex);
			if (DEBUG_THREADS) printf ("audio_pause has mutex");
			data->state = AUDIO_STATE_PLAY;
			pthread_cond_signal (&data->unpause_cond);
			pthread_mutex_unlock (&data->mutex);	/** release mutex - we're done with "data" */
		}
		else
		{
			if (DEBUG_AUDIO) report ("pausing", REP_DEBUG);
/*			while (pthread_mutex_trylock (&data->mutex) == EBUSY) { printf ("audio_pause: \n"); }*/	/** get mutex - we work on "data" */
			pthread_mutex_lock (&data->mutex);
			if (DEBUG_THREADS) printf ("audio_stop has mutex");
			data->state = AUDIO_STATE_PAUSE;
			pthread_mutex_unlock (&data->mutex);	/** release mutex - we're done with "data" */
		}
		return 1;	/* success */
	}
	data->cb_error (DAISY_ERROR_AUDIO_PAUSED_WHILE_NOT_PLAYING, "audio engine was paused while not playing");
	return -1;	/* pause failed */
}


/**
 * callback should be called by the decoder loop
 * @progress_ms the progress (so far) in the current segment. value is in milliseconds.
 * if it returns -1, the decoder should stop.
 */
int callback (long int progress_ms)
{
	int ret = 0;	/* default return value */
	void (*progress_callback) (long int);

/*	while (pthread_mutex_trylock (&data->mutex) == EBUSY) { printf ("callback retrying mutex\n"); sleep (1); }*/	/** get mutex - we work on "data" */
	pthread_mutex_lock (&data->mutex);
	if (DEBUG_THREADS) printf ("audio_callback has mutex\n");
	assert (data->magic == AUDIO_MAGIC_NUMBER);	/* check magic number */

	if (data->state == AUDIO_STATE_PAUSE)
		pthread_cond_wait (&data->unpause_cond, &data->mutex);

	if (data->state == AUDIO_STATE_STOP) {
		if (DEBUG_AUDIO) report ("callback stopping audio decoder", REP_DEBUG);
		ret = -1;	/* -1 tells the decoder to stop */
	}
	
	progress_callback = data->cb_progress;	/* copy progress callback to a local function pointer (data must be mutex locked) */
	pthread_mutex_unlock (&data->mutex);	/* release mutex - we're done with "data" */

	progress_callback (progress_ms);	/* call progress callback */
	return ret;
}


/**
 * true if the current file is difrent from
 * the last time this function was called
 * 
 * data must be under a mutex lock when this method is called!
 */
static
bool is_new_file (audio_data_t *data)
{
	assert (data->magic == AUDIO_MAGIC_NUMBER);	/* check magic number */
	
	if (data->prev_file_name != NULL)	/* prev_file_name is already set to a value */
	{
		if (!strcmp (data->prev_file_name, data->file_name))
		{
			return false;	/* the filenames are the same */
		}
		else
		{
			return true;	/* the filenames differ */
		}
	}
	else
	{
		data->prev_file_name = data->file_name;
		return true;
	}
}


/**
 * This function will load a file if it's not already loaded.
 * The freeing and generating of mmaps are handled here.
 * Returns false on error, true on success.
 * 
 * data must be mutex locked when this function is called!
 */
static
bool open_file (const char *fname, audio_data_t *data)
{
	assert (data->magic == AUDIO_MAGIC_NUMBER);	/* check magic number */

	/* did we already load the file? */
	/* We have to check if data->file_name is NULL /first!/ If one of
	 * strcmp's arguments are NULL, we've made a booboo */
	if ((data->file_name != NULL) && !strcmp (data->file_name, fname))
	{
		if (DEBUG_AUDIO) report ("File already loaded", REP_DEBUG);
	}

	/* if not, we'll need to free the old mmap, open the new
	 * file and create a new mmap */
	else	/** file is NOT already loaded */
	{
		/* Free mmap, clean data */
		if (data->fdm != NULL) {

			/* freeing the mmap before creating a new one */
			if (munmap (data->fdm, data->stat.st_size) == -1)
			{
				data->cb_error (DAISY_ERROR_AUDIO_FREE_MMAP, NULL);
				report ("Unable to free mmap", REP_ERROR);
			}

			if (DEBUG_AUDIO) {
				char rep[STRLEN];
				snprintf (rep, sizeof (rep), "Freed mmap. Pointer: %p, Size: %d", data->fdm, (int) data->stat.st_size);
				report (rep, REP_DEBUG);
			}

			data->fdm = NULL;		/* we have freed the data - don't want an old pointer to point there! */
			data->file_name = NULL;	/* this file is no longer in our memory */
			close (data->fd);		/* closing the file before we open a new one */
		}
		
		/* udating the data-struct with the new file_name */
		if (data->file_name != NULL)
			free(data->file_name);
		data->file_name = (char *) malloc (strlen(fname)+1);
		if (data == NULL) {
			data->cb_error (DAISY_ERROR_AUDIO_MALLOC, NULL);
			return false;	/* malloc failed */
		}
		strcpy (data->file_name, fname);
		
		/* opening new file */
		if ((data->fd = open (fname, O_RDONLY)) == -1)
		{
			data->cb_error (DAISY_ERROR_AUDIO_OPEN, NULL);
			report ("Error opening file", REP_WARNING);
			return false;	/* open file failed */
		}

		/* setting filetype */
		data->prev_file_type = data->file_type;
		data->file_type = get_file_type(fname);

		/* fstating file. this gives us alot of information about
		 * the file (we'll use the st_size (filesize) here) */
		if (stat(fname, &data->stat) == -1 || data->stat.st_size == 0)
		{
			data->cb_error (DAISY_ERROR_AUDIO_FSTAT, NULL);
			report ("Error fstating file or filesize is 0", REP_WARNING);
			return false;	/* fstat failed */
		}

		/* mmaps the file */
		data->fdm = mmap (0, data->stat.st_size, PROT_READ, MAP_SHARED, data->fd, 0);
		if (data->fdm == MAP_FAILED)
		{
			data->cb_error (DAISY_ERROR_AUDIO_CREATE_MMAP, NULL);	/* report error */
			report ("Error mmapping file", REP_WARNING);
			return false;	/* mmap failed */
		}
		
		if (DEBUG_AUDIO)
		{
			char rep[STRLEN];		
			snprintf (rep, sizeof (rep), "mmapped new file. Pointer: %p, size: %d", data->fdm, (int) data->stat.st_size);		
			report (rep, REP_DEBUG);
		}
	}

	return true;
}



/*
 * Finds out what filetype a file is (mp3, aac, wav).
 * TODO: Make this come up with an intelligent guess at least
 * wether the file is mp3 or mp4/aac.
 */
static
audio_file_t get_file_type (const char *fname)
{
	assert (fname);
/*	if (DEBUG_AUDIO) report("Querying file_type. Returning mp3 for now.", REP_DEBUG); */
	return AUDIO_FILE_MP3;
}


/**
 * This is called when the program wants us to stop and free our resources.
 * This should make sure resources are freed.
 */
void audio_terminate (void)
{
/*	while (pthread_mutex_trylock (&data->mutex) == EBUSY) { printf ("retrying mutex\n"); }*/	/* get mutex - we work on "data" */
	pthread_mutex_lock (&data->mutex);
	if (DEBUG_THREADS) printf ("audio_terminate has mutex\n");

	assert (data->magic == AUDIO_MAGIC_NUMBER);	/* check magic number */

	data->state = AUDIO_STATE_STOP;

	switch (data->file_type)	/* call the correct decoder terminate-function */
	{
		case (AUDIO_FILE_MP3):
			mp3_close(data);
			break;
		case (AUDIO_FILE_AAC):
#ifdef WITH_AAC
			aac_close(data);
#endif	/* WITH_AAC */
			break;
		case (AUDIO_FILE_WAV):
/*				wav_close(data); */
			break;
		default:
			break;
	}
	
	/* moved from mp3_close (). It makes more sense to have this here. */

	if (DEBUG_THREADS) printf ("mp3_close has mutex\n");

	/* Attempting to free mmap */
	if (data->fdm != NULL) {	/* if data->fdm is null, no file is read into memory and we won't have to free anything */
		if (munmap (data->fdm, data->stat.st_size) == -1) {
			data->cb_error (DAISY_ERROR_AUDIO_FREE_MMAP, NULL);
			report ("Error freeing mmap", REP_MESSAGE);
		}
	}

	close (data->fd);	/* closing the file */


	munmap (data->fdm, data->stat.st_size);	/* free the mmap'ed audio file */
	free (data->file_name); /* release space allocated for the filename */

	data->magic = 0x00000000;	/* resetting magic number - we can't use struct anymore! */

	pthread_mutex_unlock (&data->mutex);	/* release mutex - we're done with "data" */

	free (data);	/* Free data struct */
}


/**
 * This is called when the program wants us to start and prepare for audio
 * playback
 * This should allocate the resources we need.
 * 
 * @done is a callback function pointer that will be called once a segment
 * have been played.
 * @error is a callback function pointer that will be called when errors occur.
 */
int audio_initiate (daisyplayer_t daisy,
					 void (*done) (daisyplayer_t d),
					 void (*error) (enum daisy_status, char *daisy_status_msg),
					 void (*progress) (long int progress_ms))
{
	data = (audio_data_t *) malloc (sizeof (audio_data_t));
	if (data == NULL) {
		error (DAISY_ERROR_AUDIO_INITIATE_DATA, NULL);
		report ("Cannot malloc size for data", REP_ERROR);
		return -1;	/* fail */
	}

	data->magic = AUDIO_MAGIC_NUMBER;	/* set magic number */
	assert (data->magic == AUDIO_MAGIC_NUMBER);
	
	/* And setting some values to safe defaults */
	data->fdm = NULL;
	data->file_name = NULL;
	data->prev_file_name = NULL;

	data->file_type = AUDIO_FILE_UNKNOWN;
	data->prev_file_type = AUDIO_FILE_UNKNOWN;

	data->state = AUDIO_STATE_STOP;		/* default state is stopped */
	data->is_playing = false;
	
	/* callback functions */	
	data->cb_playing_done = done;	/* callback function. playback_done () should
									   be called whenever we're done playing
									   a segment */
	data->do_callback_done = false;

	data->cb_error = error;			/* callback function. cb_error () should
								be called when an error occur. */

	data->cb_progress = progress;	/* callback function. cb_progress are called during playback */

	data->daisy = daisy;	/* we keep a pointer to the main daisy struct */

	/* Setting up some thread data */
	if (pthread_cond_init(&data->unpause_cond, NULL) != 0) {
		data->cb_error (DAISY_ERROR_AUDIO_INITIATE_DATA, NULL);
		report ("Could not initialize data->unpause_cond", REP_ERROR);
		return -1; /* fail */
	}
	if (pthread_cond_init(&data->finished_cond, NULL) != 0) {
		data->cb_error (DAISY_ERROR_AUDIO_INITIATE_DATA, NULL);
		report ("Could not initialize data->finished_cond", REP_ERROR);
		return -1; /* fail */
	}
	if (pthread_cond_init(&data->stopped_cond, NULL) != 0) {
		data->cb_error (DAISY_ERROR_AUDIO_INITIATE_DATA, NULL);
		report ("Could not initialize data->stopped_cond", REP_ERROR);
		return -1; /* fail */
	}
	if (pthread_cond_init(&data->playing_cond, NULL) != 0) {
		data->cb_error (DAISY_ERROR_AUDIO_INITIATE_DATA, NULL);
		report ("Could not initialize data->playing_cond", REP_ERROR);
		return -1; /* fail */
	}
	if (pthread_mutex_init(&data->mutex, NULL) != 0) {
		data->cb_error (DAISY_ERROR_AUDIO_INITIATE_DATA, NULL);
		report ("Could not initialize data->mutex", REP_ERROR);
		return -1; /* fail */
	}

	if (DEBUG_AUDIO) report ("data struct initialized", REP_DEBUG);

	return 1;
}
