/***************************************************************************
 *   Copyright (C) 2006 by André Lindhjem <belgarat@sdf.lonestar.org>,     *
 *                         Kjetil Holien <kjetil.holien@gmail.com>,        *
 *                         Terje Risa <terje.risa@gmail.com> &             *
 *                         Øyvind Nerbråten <oyvind@nerbraten.com>         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************
 *
 * \file audio.h
 * \author André Lindhjem, Kjetil Holien, Terje Risa & Øyvind Nerbråten
 * \date 25.02.2006
 */

#ifndef AUDIO_H_
#define AUDIO_H_

#include <mad.h>		/* mad_timer_t */
#include <ao/ao.h>		/* ao_sample_format */
#include <sys/stat.h>	/* file stat */
#include <pthread.h>

#include "daisylibao.h"
#include "libdaisy.h"



/**
 * type to identify difrent audio file formats
 */
typedef enum AUDIO_FILE_TYPE 
{
	AUDIO_FILE_WAV,				/* PCM WAVE file */
	AUDIO_FILE_MP3,				/* MP3 file. uses mp3.h */
	AUDIO_FILE_AAC,				/* AAC file. uses aac.h */
	AUDIO_FILE_UNKNOWN			/* Unknown filetype */
} audio_file_t;


/**
 * message type that we send to the decoders
 */
typedef enum {
	AUDIO_STATE_PLAY,
	AUDIO_STATE_PAUSE,
	AUDIO_STATE_STOP,
	AUDIO_STATE_NOOP
} audio_state_t;


/**
 * plays a file from start seconds/fractions to stop second/fractions
 * returns the duration in milliseconds
 */
unsigned
long
int audio_play (const char *fname, const char *start, const char *stop);


/**
 * pauses the current playback
 */
int audio_pause (void);


/**
 * stops the current playback
 */
int audio_stop (void);


/**
 * callback should be called by the decoder loop
 */
int callback (long int);
 
 
/**
 * terminates audio and cleans up
 */
void audio_terminate (void);


/**
 * initiates audio and allocates data
 */
int audio_initiate (daisyplayer_t daisy,
					 void (*done) (daisyplayer_t d),
					 void (*error) (enum daisy_status, char *),
					 void (*progress) (long int progress_ms));


/**
 * the audio thread function. this should run as a thread from daisyplayer.c
 * while the audio_engine is used.
 */
void* audio_thread (void*);


/**
 * returns the currant state of the audio engine
 */
audio_state_t audio_state (void);


/**
 * This is the data. All information related to the data should be in this struct.
 */
typedef struct
{
	/* buffer */
	struct buffer
	{
		unsigned char const *buffer_start;
		unsigned long buffer_length;
	} buffer;
	
	char			*file_name;			/* the file_name of the file we have loaded */
	char			*prev_file_name;	/* the previous file_name. we check against this so that we won't have to reload already opened files */

	int				fd;				/* input file descriptor */
	void			*fdm;			/* pointer to file mmap */
	struct stat		stat;			/* file stats */

	audio_file_t	file_type;			/* the file_type (mp3, aac, wav) */
	audio_file_t	prev_file_type;		/* the previous file_type */
	
	ao_sample_format format;	/* the format of our output samples. we use this to initialize ao */

	/* mad_timer_t variables to keep track of progress, start and stop-times. */
	/* TODO: We'll have to use this for all decoders. Define our own struct in audio.h and use this instead? */
	mad_timer_t		progress;
	mad_timer_t		start;
	mad_timer_t		stop;
	
	/* method to pass messages to the decoders */
	audio_state_t state;
	bool is_playing;
	
	bool first_run;
	
	pthread_mutex_t		mutex;		/* we need to prevent race conditions on shared variables */

	pthread_cond_t		finished_cond;
	pthread_cond_t		stopped_cond;
	pthread_cond_t		playing_cond;
	pthread_cond_t		unpause_cond;
	
	/* callback functions */
	void (*cb_playing_done) (daisyplayer_t d);	/* audio done callback function */
	bool do_callback_done;	/* set to true to run cb_playing_done once */
	void (*cb_error) (enum daisy_status, char *);	/* error callback */
	void (*cb_progress) (long int progress_ms);	/* error callback */
	
	daisyplayer_t daisy;	/* this will need to be passed along with cb_playing_done (is it a good way of doing it though?) */
	
	int magic;	/* magic number */
} audio_data_t;


#endif /* AUDIO_H_ */
