/***************************************************************************
 *   Copyright (C) 2006 by André Lindhjem <belgarat@sdf.lonestar.org>,     *
 *                         Kjetil Holien <kjetil.holien@gmail.com>,        *
 *                         Terje Risa <terje.risa@gmail.com> &             *
 *                         Øyvind Nerbråten <oyvind@nerbraten.com>         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************
 *
 * \file libdaisy.c
 * \author André Lindhjem, Kjetil Holien, Terje Risa & Øyvind Nerbråten
 * \date 21.03.2006
 */
 
 
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <assert.h>

#include "libdaisy.h"
#include "control.h"
#include "common.h"
#include "snprintf/snprintf.h"
#include "audio.h"


#define DAISYPLAYER_MAGIC_NUMBER 0xC729A4BF

static void 	local_cb_daisy_audio_done	(daisyplayer_t d);


/**
 * This struct needs to be initalized and passeed along with the daisy API functions.
 * Contains pointers to the callback funtions.
 */
typedef struct {
	void		(*cb_daisy_audio_done)	(void *);	/* (const progress_ms_t);	called when libdaisy are done playing an audio segment */
	void		(*cb_daisy_audio_next)	(void *);	/* called when daisy are playing audio */
	void		(*cb_daisy_text)		(void *);	/* (const passage_t);	called when daisy have parsed a text-passage */
	void		(*cb_daisy_id)			(void *);	/* (const id_t);	called when daisy have parsed an id */
	void		(*cb_daisy_error)		(enum daisy_status, char *daisy_status_msg);
	void		(*cb_daisy_progress)	(long int);	/* (long int progress_ms) */
	
	int magic;
} struct_daisyplayer_t;


/**
 * This functions initializes the daisy library.
 * It should be called before any atempt to use daisy functionality.
 * @param void l_cb_daisy_audio_done - this function will be called when an audio segment is done playing.
 * @param void l_cb_daisy_audio_next - this function will be called when an new audio segment starts playing, supplying the duration of the segment in ms.
 * @param void l_cb_daisy_text - this function will be called when an new audio segment starts playing, supplying the text corresponding to the audio.
 * @param void l_cb_daisy_id - this function will be called when an new audio segment starts playing, supplying the id of the text passage in the xml file.
 * @param void l_cb_daisy_error - this function will be called when an engine error occures.
 * @param void l_cb_daisy_progress - this function will be called during playback, supplying the progress in ms.
 * @returns daisyplayer_t - the daisy data stuct which must be passed along with all the API functions.
 */
daisyplayer_t daisy_init (void (*l_cb_daisy_audio_done)	(void *),			/* (const progress_ms_t ), */
						  void (*l_cb_daisy_audio_next)	(void *),			/* (void), */
						  void (*l_cb_daisy_text)		(void *),			/* (const passage_t ), */
						  void (*l_cb_daisy_id)			(void *),			/* (const id_t ) */
						  void (*l_cb_daisy_error)		(enum daisy_status, char *),
						  void (*l_cb_daisy_progress)	(long int))			/* (long int progress_ms) */
						  
{
	struct_daisyplayer_t *daisy = malloc (sizeof(struct_daisyplayer_t));	/* allocates space for struct daisyplayer */

	daisy->magic = DAISYPLAYER_MAGIC_NUMBER;	/* set magic number */

	/* initiates the callback functions */
	daisy->cb_daisy_audio_done	= l_cb_daisy_audio_done;
	daisy->cb_daisy_audio_next	= l_cb_daisy_audio_next;
	daisy->cb_daisy_text		= l_cb_daisy_text;
	daisy->cb_daisy_id			= l_cb_daisy_id;
	daisy->cb_daisy_error		= l_cb_daisy_error;
	daisy->cb_daisy_progress	= l_cb_daisy_progress;
	
	/* initiates the parser and data structure */
	parseInit ();
	
	/* initiate the audio thread */
	start_audio (daisy, local_cb_daisy_audio_done, daisy->cb_daisy_error, daisy->cb_daisy_progress);
	
	return daisy;
}


/**
 * This functions terminates the daisy library.
 * It should be called when daisy are no longer needed.
 * @param daisy - the daisy data stuct which must be passed along with all the API functions.
 */
void daisy_term (daisyplayer_t daisy)
{
	struct_daisyplayer_t *_daisy	= daisy;
	
	parseTerminate ();	/* deallocates the data structure and cleans up after the parser */
	
	audio_terminate();	/* free resources used by the audio engine */
	
	_daisy->magic = 0x00000000;
	free (_daisy);		/* frees the daisy struct */
}


/**
 * Loades a new daisy book ready for playback
 * @param daisy - the daisy data stuct which must be passed along with all the API functions.
 * @param path - the path to the daisy dtb (ncc.* | *.ncx)
 * @returns 1 if success and -1 if not
 */
int daisy_load (daisyplayer_t daisy, char *path)
{
	int result;
	
	daisy = NULL;
	result = parse (path);
	if (result != 1)
		return result;
	result = seek (DAISY_SEEK_NEXT_PASSAGE);
	return result;
}

/**
 * Function to start playback.
 * @param daisy - the daisy data stuct which must be passed along with all the API functions.
 * @returns 1 on success, and always succeeds.
 */
int daisy_play (daisyplayer_t daisy)
{
	struct_daisyplayer_t *_daisy = daisy;
	unsigned long int *duration_ms;
	char audiopath[STRLEN];
	
	/* check if a daisy dtb is loaded */
	if (__smilHead->next == __smilTail)
	{
		_daisy->cb_daisy_error (DAISY_ERROR_PLAYBACK_NO_DTB_LOADED, NULL);
		return -1;	
	}

	/* starting audio */
	if (__nodePos->audiofilename != NULL)
	{
		snprintf (audiopath, sizeof (audiopath), "%s%s", __path, __nodePos->audiofilename);
		duration_ms = (unsigned long int *)malloc (sizeof (unsigned long int));
		*duration_ms = audio_play (audiopath, __nodePos->audioStartPos, __nodePos->audioStopPos);
		_daisy->cb_daisy_audio_next (duration_ms);	/* we start playing a passage */
	}
	else
	{
		_daisy->cb_daisy_error (DAISY_ERROR_PLAYBACK_NO_AUDIO_IN_SEGMENT, NULL);
	}

	/* reporting id and text */
	if (__nodePos->textPassage != NULL)
	{
		_daisy->cb_daisy_id ((void *) __nodePos->fragmentIdentifier);	/* output id */
		_daisy->cb_daisy_text ((void *) __nodePos->textPassage);	/* output text */
	}
	else
	{
		_daisy->cb_daisy_error (DAISY_ERROR_PLAYBACK_NO_TEXT_IN_SEGMENT, NULL);	
	}
	
	return 1;	/* success */
}


/**
 * Function to seek during playback.
 * @param daisy - the daisy data stuct which must be passed along with all the API functions.
 * @param seek_option: an daisy_seek_option.
 * @returns 0 if end_of_book, 1 if success and -1 on error
 */
int daisy_seek (daisyplayer_t daisy, int seek_option)
{	
	struct_daisyplayer_t *_daisy = daisy;
	int result;
	
	/* check if a daisy dtb is loaded */
	if (__smilHead->next == __smilTail)
	{
		_daisy->cb_daisy_error (DAISY_ERROR_PLAYBACK_NO_DTB_LOADED, NULL);
		return -1;	
	}
	
	daisy_stop (daisy);
	result = seek (seek_option);
	if (result != 1)
		return result;
	daisy_play(daisy);
	return result;	
}


/**
 * Function to stop playback.
 * @param daisy - the daisy data stuct which must be passed along with all the API functions.
 * @returns 1 on success. (and always succeeds)
 */
int daisy_stop (daisyplayer_t daisy)
{
	daisy = NULL;	/* daisy struct is unused */
	return audio_stop ();	/* audio_stop is void and always succeeds */
}

/**
 * Function to pause playback.
 * @param daisy - the daisy data stuct which must be passed along with all the API functions.
 * @returns 1 on success, -1 on fail.
 */
int daisy_pause (daisyplayer_t daisy)
{
	daisy = NULL;	/* daisy struct is unused */
	return audio_pause ();
}

/**
 * Internal callback function. Will be called each time the audio engine
 * is done playing the supplied segment. Feeds the audio engine
 * with the next audio segment and signals the the front end that an
 * segment is done playing.
 * @param daisy - the daisy data stuct which must be passed along with all the API functions.
 */
static void local_cb_daisy_audio_done (daisyplayer_t daisy)
{
	int result;
	struct_daisyplayer_t *_daisy = daisy;
	
	_daisy->cb_daisy_audio_done (NULL);

	result = seek (DAISY_SEEK_NEXT_PASSAGE);	/* one passage is done - seeking to the next passage */
	if (result != 1 )
	{
		if (result == 0)	_daisy->cb_daisy_error (DAISY_END_OF_BOOK, NULL);
		else				_daisy->cb_daisy_error (DAISY_ERROR_UNKNOWN, NULL);
	}
	else
	{
		daisy_play (daisy);
	}
	
}
