/***************************************************************************
 *   Copyright (C) 2006 by André Lindhjem <belgarat@sdf.lonestar.org>,     *
 *                         Kjetil Holien <kjetil.holien@gmail.com>,        *
 *                         Terje Risa <terje.risa@gmail.com> &             *
 *                         Øyvind Nerbråten <oyvind@nerbraten.com>         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************
 *
 * \file nccparser.c
 * \author André Lindhjem, Kjetil Holien, Terje Risa & Øyvind Nerbråten
 * \date 16.02.2006
 *
 * NCC parser for the Daisy 2.02 standar. Collects data and stores it in linked list.
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libxml2/libxml/xmlreader.h>

#include "common.h"
#include "nccparser.h"
#include "control.h"
#include "report.h"
#include "snprintf/snprintf.h"

#ifdef LIBXML_READER_ENABLED

/* static function declarations */
static int parseNccHead (xmlTextReaderPtr reader);
static int parseNccTitle (xmlTextReaderPtr reader);
static int parseNccMeta (xmlTextReaderPtr reader);
static int parseNccBody (xmlTextReaderPtr reader);
static int parseNccHx (xmlTextReaderPtr reader, struct SmilNode *tmpSmilNode);
static int parseNccSpan (xmlTextReaderPtr reader, struct SmilNode *tmpSmilNode);
static int parseNccDiv (xmlTextReaderPtr reader, struct SmilNode *tmpSmilNode);
static int parseNccA (xmlTextReaderPtr reader, struct SmilNode *tmpSmilNode);

/* global variables */
extern char *path;


/************************ Code for parsing a NCC.* file ************************/


/**
 * Method for parsing of a NCC.* file
 * 
 * @param filname - the name of the ncc file
 * @returns 1 in case of success and -1 in case of error
 */
int parseNCC (char *filename)
{
 	const xmlChar *name = NULL;
 	xmlTextReaderPtr reader = NULL;
 	int ret = 1;
 	
 	/* open the ncc file */
 	reader = xmlReaderForFile (filename, NULL, 0);
 	if (reader == NULL)
 	{
 		char error[STRLEN];
 		snprintf (error, sizeof(error), "Failed to open NCC file: %s", filename);
 		report (error, REP_WARNING);
 		return -1;
 	}
 	else 
	{
		if (DEBUG_NCCPARSER)
		{
			char error[STRLEN];
			snprintf (error, sizeof(error), "Parsing NCC file: %s", filename);
	 		report (error, REP_DEBUG);
		}
	 	/* read the first tag */
	 	ret = xmlTextReaderRead(reader);
	 	if (ret != 1)
	 	{
	 		char error[STRLEN];
	 		snprintf (error, sizeof(error), "Failed to parse NCC file: %s", filename);
 			report (error, REP_WARNING);
 			xmlFreeTextReader (reader);
 			return -1;
	 	}
	 	/* while there are more tags, parse the next */
	 	while(ret == 1)
	 	{
	 		/* get name of tag */
			name = xmlTextReaderConstName (reader);
			if (name == NULL) return 0;
			/* parse tag if <head> */
			if (!strcmp ((const char *)name, "head"))
			{ 
				ret = parseNccHead (reader);
				if (ret != 1)
			 	{
			 		char error[STRLEN];
			 		snprintf (error, sizeof(error), "Failed to parse NCC file: %s", filename);
		 			report (error, REP_WARNING);
		 			xmlFreeTextReader (reader);
		 			return -1;
			 	}
			}
			/* parse tag if <body> */
			else if (!strcmp ((const char *)name, "body"))
			{ 
				ret = parseNccBody (reader);
				if (ret != 1)
			 	{
			 		char error[STRLEN];
			 		snprintf (error, sizeof(error), "Failed to parse NCC file: %s", filename);
		 			report (error, REP_WARNING);
		 			xmlFreeTextReader (reader);
		 			return -1;
			 	}
			}
			ret = xmlTextReaderRead (reader);
		}
		xmlFreeTextReader (reader);
		if (ret != 0)
		{
			char error[STRLEN];
			snprintf (error, sizeof(error), "Failed to parse NCC file: %s", filename);
 			report (error, REP_WARNING);
 			return -1;
		}
	}
	return 1;
}
 

/**
 * Parse Head-tag.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNccHead (xmlTextReaderPtr reader)
{
	const xmlChar *name = NULL;
	int ret, type;
	ret = type = 1;
	
	if (DEBUG_NCCPARSER) report ("Parsing NCC <HEAD>", REP_DEBUG);
	type = xmlTextReaderNodeType(reader);
	if (type != 1) 
	{   if (DEBUG_NCXPARSER) report ("Failed NCC <HEAD>", REP_ERROR);
		return -1;
	}
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	if (name == NULL) return 0;
	/* while we haven't come to the head closetag */
	while (strcmp ((const char *)name, "head"))
	{	
		/* parse tag if <title> */
		if (!strcmp ((const char *)name, "title")) ret = parseNccTitle (reader);
		/* parse tag if <meta> * */
		else if (!strcmp ((const char *)name, "meta")) ret = parseNccMeta (reader);
		if (ret != 1) return ret;
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		name = xmlTextReaderConstName (reader);
		if (name == NULL) return 0;
	}	
	type = xmlTextReaderNodeType(reader);
	if (type != 15) 
	{   if (DEBUG_NCXPARSER) report ("Failed parsing NCC <HEAD>, endtag expected", REP_ERROR);
		return -1;
	}	
	return ret;
}


/**
 * Parse Title-tag. Extracts the title.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNccTitle (xmlTextReaderPtr reader)
{
	xmlChar *value = NULL;
	char *tmp = NULL;
	int ret = 1;
	
	if (DEBUG_NCCPARSER) report ("Parsing NCC <TITLE>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	/* if the title tag has a value	 */
	if (xmlTextReaderHasValue (reader))
	{
		value = xmlTextReaderValue (reader);
		tmp = removewhitespaces ((char *)value);
		/* storing the books title in the datastucture */
		__bookInfo->titleText = tmp;
	}
	if (value != NULL) xmlFree (value); value = NULL;
	ret = xmlTextReaderRead (reader);
	return ret;
}


/**
 * Parse Meta-tag. Currently not implemented.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNccMeta (xmlTextReaderPtr reader)
{
	int ret = 1;
	
	if (DEBUG_NCCPARSER) report ("Parsing NCC <META>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	return ret;	
}


/**
 * Parse Body-tag.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNccBody (xmlTextReaderPtr reader)
{
	const xmlChar *name = NULL;
	int ret, type;
	ret = type = 1;
	
	if (DEBUG_NCCPARSER) report ("Parsing NCC <BODY>", REP_DEBUG);
	type = xmlTextReaderNodeType(reader);
	if (type != 1) 
	{   if (DEBUG_NCXPARSER) report ("Failed NCC <BODY>", REP_ERROR);
		return -1;
	}
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	if (name == NULL) return 0;
	/* while we haven't come to the body closetag */
	while (strcmp ((const char *)name, "body"))
	{	
		/* parse tag if <h1> + <h2-h6> * */
		if (!strcmp ((const char *)name, "h1") ||  !strcmp ((const char *)name, "h2") || !strcmp ((const char *)name, "h3") ||  !strcmp ((const char *)name, "h4") 
		 || !strcmp ((const char *)name, "h5") ||  !strcmp ((const char *)name, "h6"))
        {
        	struct SmilNode *tmpSmilNode = addNewSmilNode();
		 	ret = parseNccHx (reader, tmpSmilNode);
			if (ret != 1) return ret;
		}
		/* parse tag if <span> * */
		else if (!strcmp ((const char *)name, "span")) 
		{
			struct SmilNode *tmpSmilNode = addNewSmilNode();
			ret = parseNccSpan (reader, tmpSmilNode);
			if (ret != 1) return ret;
		}
		/* parse tag if <div> * */
		else if (!strcmp ((const char *)name, "div"))
		{
			struct SmilNode *tmpSmilNode = addNewSmilNode();
			ret = parseNccDiv (reader, tmpSmilNode);
			if (ret != 1) return ret;
		}
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		name = xmlTextReaderConstName (reader);
		if (name == NULL) return 0;
	}	
	type = xmlTextReaderNodeType(reader);
	if (type != 15) 
	{   if (DEBUG_NCXPARSER) report ("Failed parsing NCC <BODY>, endtag expected", REP_ERROR);
		return -1;
	}	
	return ret;
}


/**
 * Parse Hx-tag.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNccHx(xmlTextReaderPtr reader, struct SmilNode *tmpSmilNode)
{
/*TODO: Hx must contain (attributes) id, h1 must contain class, h2-6 might contain class. Not taken care of now */
	const xmlChar *name = NULL;
	xmlChar *attrvalue = NULL;
	int ret = 1;
	
	if (DEBUG_NCCPARSER) report ("Parsing NCC <Hx>", REP_DEBUG);
	/* getting value if the id attribute */
	attrvalue = xmlTextReaderGetAttribute (reader, (xmlChar *)"id");
	/* storing the value of the id attribute in the smilNode datastructure */
	if (attrvalue != NULL) tmpSmilNode->id = (char *)attrvalue;
	/* fetching next tag */
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	/* getting the name of the tag */
	name = xmlTextReaderConstName (reader);
	if (name == NULL) return 0;
	/* parse tag if <a> else return error */
	if (!strcmp((const char *)name, "a"))
	{
		ret = parseNccA (reader, tmpSmilNode);
	}
	else
		return -1;
	return ret;
}


/**
 * Parse Span-tag.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNccSpan(xmlTextReaderPtr reader, struct SmilNode *tmpSmilNode)
{
/* Span must contain id and class attributes */
	const xmlChar *name = NULL;
	xmlChar *attrvalue = NULL;
	int ret = 1;
	
	if (DEBUG_NCCPARSER) report ("Parsing NCC <SPAN>", REP_DEBUG);
	/* getting value if the id attribute */
	attrvalue = xmlTextReaderGetAttribute (reader, (xmlChar *)"id");
	/* storing the value of the id attribute in the smilNode datastructure */
	if (attrvalue != NULL) tmpSmilNode->id = (char *)attrvalue;
	/* fetching next tag */
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	if (name == NULL) return 0;
	/* parse tag if <a> else return error */
	if (!strcmp((const char *)name, "a"))
	{
		ret = parseNccA (reader, tmpSmilNode);
	}
	else
		return -1;
	return ret;
}


/**
 * Parse Div-tag.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNccDiv(xmlTextReaderPtr reader, struct SmilNode *tmpSmilNode)
{
/* Span must contain id and class attributes */
	const xmlChar *name = NULL;
	xmlChar *attrvalue = NULL;
	int ret = 1;
	
	if (DEBUG_NCCPARSER) report ("Parsing NCC <DIV>", REP_DEBUG);
	/* getting value if the id attribute */
	attrvalue = xmlTextReaderGetAttribute (reader, (xmlChar *)"id");
	/* storing the value of the id attribute in the smilNode datastructure */
	if (attrvalue != NULL) tmpSmilNode->id = (char *)attrvalue;
	/* fetching next tag */
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	if (name == NULL) return 0;
	/* parse tag if <a> else return error */
	if (!strcmp((const char *)name, "a"))
	{
		ret = parseNccA (reader, tmpSmilNode);
	}
	else
		return -1;
	return ret;
}


/**
 * Parse A-tag. Extracts the source and anchor.
 *
 * @param reader XmlTextReader.
 * @param tmpSmilNode - the smilNode to put the parsed information in.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNccA (xmlTextReaderPtr reader, struct SmilNode *tmpSmilNode)
{
	xmlChar *smilsrc, *tmp, *value;
	char *tempChar = NULL;
	int ret = 1;
	smilsrc = tmp = value = NULL;
	
	if (DEBUG_NCCPARSER) report ("Parsing NCC <A>", REP_DEBUG);
	
	/* gets the href attribute. */
	smilsrc = xmlTextReaderGetAttribute (reader, (xmlChar *)"href");	/*TODO: what if attribute does not exist? */
	
	/* splits the text on #. */
	tmp = (xmlChar *)strtok ((char *)smilsrc, "#");
	tempChar = (char *) malloc (strlen ((char *)tmp)+1);
	strcpy(tempChar, (char *)tmp);
	
	if (DEBUG_NCCPARSER)
	{
		char error[STRLEN];
		snprintf (error, sizeof(error), "Parsing NCC <A> File anchor: %s", tmp);
		report (error, REP_DEBUG);
	}
	
	/* put the anchor in the node */
	tmpSmilNode->anchor = tempChar; tempChar = NULL; 
	/* get the fragment identifier */
	tmp = (xmlChar *)strtok (NULL, "#");
	tempChar = (char *) malloc (strlen ((char *)tmp)+1);
	strcpy(tempChar, (char *)tmp);
	
	if (DEBUG_NCCPARSER)
	{
		char error[STRLEN];
		snprintf (error, sizeof(error), "Parsing NCC <A> Fragment Identifier: %s", tmp);
		report (error, REP_DEBUG);
	}
	
	/* put the fragment identifier in the node */
	tmpSmilNode->fragmentIdentifier = tempChar; tempChar = NULL;
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	
	if (smilsrc != NULL) xmlFree (smilsrc); smilsrc = NULL;
	
	/* get the text value of the node */
	if (xmlTextReaderHasValue (reader))
	{
		value = xmlTextReaderValue (reader);
		tempChar = removewhitespaces ((char *)value);
		
		if (DEBUG_NCCPARSER)
		{
			char error[STRLEN];
			snprintf (error, sizeof(error), "Parsing NCC <A> Heading: %s", value);
			report (error, REP_DEBUG);
		}
		/* put the text in the node */
		tmpSmilNode->header = tempChar; tempChar = NULL;
		if (value != NULL) xmlFree (value); value = NULL;
	}
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	ret = xmlTextReaderRead (reader);
	return ret;
}


/************************ END of NCC.* parser code ************************/


#endif
